from Crypto.Cipher import AES
from Crypto import Random
from Crypto.Hash import SHA256, SHA512
import base64

bs = AES.block_size

#
# @param key: password in python string type
# @return: SHA-512 hash output in python string type.
# 
def hash(key):
    sha512 = SHA512.new(key.encode())
    return sha512.hexdigest()

#
# @param key: password in python string type.
# @return: SHA-256 in byte string type.
# 
def hashEnc(key):
    sha256 = SHA256.new(key.encode())
    return sha256.digest()

# Mainly used by encrypt function.
def pad(s):
    return s + (bs - len(s) % bs) * chr(bs - len(s) % bs)

# Mainly used by decrypt function.
def unpad(s):
    return s[:-ord(s[len(s)-1:])]

#
# @param key: SHA-256 byte string of primary password.
# @param raw: Data to be encrypted.
# @return: Encrypted data.
# Encryption used: AES-256. 
# 
def encrypt(raw, key):
    raw = pad(raw)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode()

#
# @param key: SHA-256 byte string of primary password.
# @param enc: Encrypted data.
# @return: Decrypted data.
#
def decrypt(enc, key):
    enc = base64.b64decode(enc.encode())
    iv = enc[:AES.block_size]
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(enc[AES.block_size:])).decode('utf-8')

    


'''
1. Connect to the database
2. Create the tables
3. Insert primary password
'''

import sqlite3 
from os import environ, mkdir, path
from sys import platform
from .Encryption import hash

fileName = "Passwords.db"
pathName = None
if platform.lower().startswith("win"):
    pathName = f"{environ['APPDATA']}\\SimplePasswordManager"
else:
    pathName = f"{environ['HOME']}/.config/SimplePasswordManager" 
if not path.exists(pathName):    
    mkdir(pathName)
fileName = path.join(pathName, fileName)

def setupDatabase(primaryPassword):
    conn = sqlite3.connect(fileName)
    cur = conn.cursor()

    # Create the table 
    createTableSyntax = [
        '''
        CREATE TABLE "Passwords" (
	        "email"	TEXT,
	        "website"	TEXT,
        	"password"	TEXT,
        	PRIMARY KEY("email","website")
        );
        ''',
        '''
        CREATE TABLE "PrimaryPassword" (
	        "password"	TEXT
        );
        '''
    ]

    for syntax in createTableSyntax:
        cur.execute(syntax)
        conn.commit()

    # Insert the primary password

    syntax = f"INSERT INTO PrimaryPassword('password') VALUES('{hash(primaryPassword)}');"
    cur.execute(syntax)
    conn.commit()

    cur.close()
    conn.close()



'''
1. Do the login
2. If the login is successful, use the working class
3. Else, no access allowed.
'''

import sqlite3
from .setup import fileName 
from .Encryption import *

class Working:
    def __init__(self, primaryPassword: str):
        self.conn = sqlite3.connect(fileName)
        self.cur = self.conn.cursor()

        self.cur.execute("SELECT password FROM PrimaryPassword;")
        hashFromDatabase = self.cur.fetchall()[0][0]

        if hashFromDatabase == hash(primaryPassword):
            # Login successful.
            self._privateKey = hashEnc(primaryPassword)
            self.loginStatus = True # True means successful login.
        else:
            self.loginStatus = False
            self._privateKey = None 

    def changePrimaryPassword(self, oldPassword: str, newPassword: str) -> bool:
        self.cur.execute("SELECT password FROM PrimaryPassword;")
        hashFromDatabase = self.cur.fetchall()[0][0]

        if hashFromDatabase == hash(oldPassword):
            # Can change password.
            self.cur.execute(f"UPDATE PrimaryPassword SET password = '{hash(newPassword)}' WHERE password = '{hash(oldPassword)}';")
            self.conn.commit()
            entries = self.seeEntries()
            newKey = hashEnc(newPassword)

            for entry in entries:
                self.cur.execute(f"UPDATE Passwords SET password = '{encrypt(self.getEntry(entry[0], entry[1]), newKey)}' WHERE (email = '{entry[0]}' AND website = '{entry[1]}');")
                self.conn.commit()
            
            self._privateKey = newKey
            return True
        else:
            return False 

    def getEntry(self, email: str, website: str) -> str:
        if self.loginStatus:    
            try:    
                self.cur.execute(f"SELECT password FROM Passwords WHERE (email = '{email}' AND website = '{website}');")
                encryptedData = self.cur.fetchall()[0]
                return decrypt(encryptedData[0], self._privateKey)
            except Exception:
                return ""
        return ""

    def putEntry(self, email: str, website: str, password: str) -> bool:
        if self.loginStatus:
            try:
                self.cur.execute(f"INSERT INTO Passwords (email, website, password) VALUES ('{email}', '{website}', '{encrypt(password, self._privateKey)}');")
                self.conn.commit()
                return True
            except Exception as e:
                return False
        return False

    def updateEntry(self, oldEmail: str, oldWebsite: str, email: str, website: str, password: str) -> None:
        if self.loginStatus:    
            self.cur.execute(f"UPDATE Passwords SET email = '{email}', website = '{website}', password = '{encrypt(password, self._privateKey)}' WHERE (email = '{oldEmail}' AND website = '{oldWebsite}');")
            self.conn.commit()

    def seeEntries(self):
        if self.loginStatus:    
            self.cur.execute("SELECT email, website FROM Passwords;")
            return self.cur.fetchall()
            # To return in format of (email, website) tuple.
        else:
            return []

    def __del__(self):
        self.loginStatus = False
        self.cur.close()
        self.conn.close()